{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzAzim;

{**********************************************************************
NAME                             AZIMUTHAL EQUIDISTANT
PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Azimuthal Equidistant projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.
PROGRAMMER              DATE
----------              ----
T. Mittan		Mar, 1993
ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
#include <stdio.h>
#include "gctpc/cproj.h"
#include "gctpc/report.h"}

interface
uses Math;

//* Variables common to all subroutines in this code file
var
  r_major, //* major axis 				*/
    lon_center, //* Center longitude (projection center) */
    lat_origin, //* center latitude			*/
    e0, e1, e2, e3, //* eccentricity constants		*/
    e, es, esp, //* eccentricity constants		*/
    false_northing, //* y offset in meters			*/
    false_easting, //* x offset in meters			*/
    sin_p12, //* sin of center latitude		*/
    cos_p12: Double; //* cos of center latitude		*/


function
  azimforint(r_maj, center_lon, center_lat,
  false_east, false_north: double): Integer;

function
  azimfor(lon, lat: double; var x, y: double): Integer;

function
  aziminvint(r_maj, center_lon, center_lat,
  false_east, false_north: double): Integer;

function
  aziminv(x, y: double; var lon, lat: double): Integer;

implementation
uses buzGlobal;

function
  azimforint(r_maj, center_lon, center_lat,
  false_east, false_north: double): Integer;
{ Initialize the Azimuthal projection
long azimforint(
double r_maj,/* major axis
 double center_lon,/* center longitude
 double center_lat,/* center latitude
double false_east, /* x offset in meters
double false_north)/* y offset in meters}
{var
 temp:double;}//* temporary variable		*/
{double e0fn(double x),e1fn(double x),e2fn(double x),e3fn(double x);
 /* 	functions	*/}
begin
//* Place parameters in static storage for common use
  r_major := r_maj;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;
  sin_p12 := Sin(center_lat);
  cos_p12 := Cos(center_lat);
  {sincos(center_lat, sin_p12, cos_p12);}

//* Report parameters to the user
  ptitle('AZIMUTHAL EQUIDISTANT');
  radius(r_major);
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  azimfor(lon, lat: double; var x, y: double): Integer;
{Azimuthal forward equations--mapping lat,long to x,y
long azimfor(
double lon,/* (I) Longitude
double lat,/* (I) Latitude
double *x,/* (O) X projection coordinate
double *y)/* (O) Y projection coordinate}
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180 	*/}
var
  sinphi, cosphi, //* sin and cos value
    dlon, //* delta longitude value
    coslon, //* cos of longitude
    ksp, //* scale factor
    g,
    con, //* radius of circle
    z: double; //* angle
{char mess[80];	}//* error message buffer
begin
//* Forward equations  -----------------*/
  dlon := AdjustLon(lon - lon_center);
  sinphi := Sin(lat);
  cosphi := Cos(lat);
  {sincos(lat, sinphi, cosphi);}
  coslon := cos(dlon);
  g := sin_p12 * sinphi + cos_p12 * cosphi * coslon;
  if (abs(abs(g) - 1.0) < EPSLN) then
  begin
    ksp := 1.0;
    if (g < 0.0) then
    begin
      con := 2.0 * HALF_PI * r_major;
      if con > 2 * con then ; {compile noise}
{     sprintf(mess,"Point projects into a circle of radius = %12.2lf",con);}
{      PError('mess', 'azim-for');}
      DoProjError(123,123,0,0,con);
      result := (123);
      exit;
    end;
  end
  else
  begin
    z := arccos(g);
    ksp := z / sin(z);
  end;
  x := false_easting + r_major * ksp * cosphi * sin(dlon);
  y := false_northing + r_major * ksp * (cos_p12 * sinphi - sin_p12 *
    cosphi * coslon);
  result := (OK);
end;

function
  aziminvint(r_maj, center_lon, center_lat, false_east,
  false_north: double): Integer;
{ Initialize the Azimuthal projection
long aziminvint(
 double r_maj,/* major axis
 double center_lon,/* center longitude
 double center_lat,/* center latitude
 double false_east,/* x offset in meters
 double false_north)/* y offset in meters}
begin
{ Place parameters in static storage for common use}
  r_major := r_maj;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;
  sin_p12 := Sin(center_lat);
  cos_p12 := Cos(center_lat);
  {sincos(center_lat, sin_p12, cos_p12);}

//* Report parameters to the user
  ptitle('AZIMUTHAL EQUIDISTANT');
  radius(r_major);
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  aziminv(x, y: double; var lon, lat: double): Integer;
{Azimuthal inverse equations--mapping x,y to lat/long
long aziminv(
double x,/* (O) X projection coordinate
double y, /* (O) Y projection coordinate
double *lon,/* (I) Longitude
double *lat)/* (I) Latitude}
{double adjust_lon(double x);
 /* Function to adjust longitude to -180 - 180 	*/
double asinz(double con);}
var
  rh, //* height above ellipsoid
    z, //* angle
    sinz, cosz, //* sin of z and cos of z
    temp,
    con: double;
begin
  result := (OK);
//* Inverse equations  -----------------*/
  x := x - false_easting;
  y := y - false_northing;
  rh := sqrt(x * x + y * y);
  if (rh > (2.0 * HALF_PI * r_major)) then
  begin
    {PError('Input data error', 'azim-inv');}
    DoProjError(125,125,0,0,0);
    result := (125);
    exit;
  end;
  z := rh / r_major;
  sinz := Sin(z);
  cosz := Cos(z);
  {sincos(z, sinz, cosz);}
  lon := lon_center;
  if (abs(rh) <= EPSLN) then
  begin
    lat := lat_origin;
    result := (OK);
    exit
  end;
  lat := asinz(cosz * sin_p12 + (y * sinz * cos_p12) / rh);
  con := abs(lat_origin) - HALF_PI;
  if (abs(con) <= EPSLN) then
  begin
    if (lat_origin >= 0.0) then
    begin
      lon := AdjustLon(lon_center + arctan(x / -y));
      result := (OK);
      exit;
    end
    else
    begin
      lon := AdjustLon(lon_center - arctan(-x / y));
      result := (OK);
    end;
  end;
  con := cosz - sin_p12 * sin(lat);

  if ((abs(con) < EPSLN) and (abs(x) < EPSLN)) then
    result := (OK);
  temp := arctan((x * sinz * cos_p12) / (con * rh));
  lon := AdjustLon(lon_center + temp);


end;



end.
