{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzEqui;
{**********************************************************************
NAME                             EQUIRECTANGULAR
PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Equirectangular projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.
PROGRAMMER              DATE
----------              ----
T. Mittan		Mar, 1993
ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}

interface
uses Math;

//* Variables common to all subroutines in this code file
var
  r_major, //* major axis
    lon_center, //* Center longitude (projection center)
    lat_origin, //* center latitude
    false_northing, //* y offset in meters
    false_easting: Double; //* x offset in meters

function
  equiforint(r_maj, center_lon, lat1, false_east,
  false_north: double): Integer;

function
  equifor(lon, lat: double; var x, y: double): Integer;

function
  equiinvint(r_maj, center_lon, lat1,
  false_east, false_north: double): Integer;

function
  equiinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

function
  equiforint(r_maj, center_lon, lat1, false_east,
  false_north: double): Integer;
{Initialize the Equirectangular projection
long equiforint(
double r_maj,/* major axis
double center_lon,/* center longitude
double lat1,/* latitude of true scale
 double false_east,/* x offset in meters
 double false_north)/* y offset in meters}
begin
//* Place parameters in static storage for common use
  r_major := r_maj;
  lon_center := center_lon;
  lat_origin := lat1;
  false_northing := false_north;
  false_easting := false_east;

//* Report parameters to the user
  ptitle('EQUIRECTANGULAR');
  radius(r_major);
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  equifor(lon, lat: double; var x, y: double): Integer;
{Equirectangular forward equations--mapping lat,long to x,y
long equifor(double lon, double lat, double *x, double *y)
              /* (I) Longitude
              /* (I) Latitude
             /* (O) X projection coordinate
             /* (O) Y projection coordinate}
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180 	*/}
var
{ sinphi, cosphi, }//* sin and cos value
  dlon //* delta longitude value
 { coslon,}//* cos of longitude
 { ksp,  }//* scale factor
{ g}: double;
begin
//* Forward equations
  dlon := AdjustLon(lon - lon_center);
  x := false_easting + r_major * dlon * cos(lat_origin);
  y := false_northing + r_major * lat;
  result := (OK);
end;



function
  equiinvint(r_maj, center_lon, lat1,
  false_east, false_north: double): Integer;
{Initialize the Equirectangular projection
long equiinvint(double r_maj, double center_lon,
double lat1, double false_east, double false_north)
                /* major axis
                    /* center longitude
               /* latitude of true scale
                    /* x offset in meters
                     /* y offset in meters}
{var
 temp:double; }//* temporary variable		*/
begin
//* Place parameters in static storage for common use
  r_major := r_maj;
  lon_center := center_lon;
  lat_origin := lat1;
  false_northing := false_north;
  false_easting := false_east;

//* Report parameters to the user
  ptitle('EQUIRECTANGULAR');
  radius(r_major);
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;

function
  equiinv(x, y: double; var lon, lat: double): Integer;
{Equirectangular inverse equations--mapping x,y to lat/long
  ---------------------------------------------------------*/
long equiinv(double x, double y, double *lon, double *lat)
            /* (O) X projection coordinate
            /* (O) Y projection coordinate
               /* (I) Longitude
               /* (I) Latitude}
{double adjust_lon(double x);
 /* Function to adjust longitude to -180 - 180 	*/}
{var
 sinphi, cosphi,	//* sin and cos value
 dlon,		//* delta longitude value
 coslon,		//* cos of longitude
 ksp,		//* scale factor
 g:double; }
begin
//* Inverse equations  -----------------*/
  x := x - false_easting;
  y := y - false_northing;
  lat := y / r_major;
  if (abs(lat) > HALF_PI) then
  begin
    {PError('Input data error', 'equi-inv'); }
    DoProjError(174,174,0,0,0);
    result := (174);
    exit;
  end;
  lon := AdjustLon(lon_center + x / (r_major * cos(lat_origin)));
  result := (OK);
end;



end.
