{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzImolw;

{**********************************************************************
NAME                        INTERRUPTED MOLLWEIDE

PURPOSE:        Transforms input longitude and latitude to Easting and
                Northing for the Interrupted Mollweide projection.  The
                longitude and latitude must be in radians.  The Easting
                and Northing values will be returned in meters.

PROGRAMMER              DATE		REASON
----------              ----		------
D. Steinwand, EROS      June, 1991	Initial development
T. Mittan, EDC		Feb, 1993	Adapted format to be used by the "C"
     version of GCTP.
S. Nelson, EDC		June, 1993	Changed precisian of radians in
     assigning the region, and in the
     conversion algorithm.

ALGORITHM REFERENCES

1.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.

2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface
uses Math;
//* Variables common to all subroutines in this code file
var
  RadiusES: Double; //* Radius of the earth (sphere) */
  lon_center: array[0..6] of double;
    //* Central meridians, one for each region */
  feast: array[0..6] of double;
    //* False easting, one for each region */

function
  imolwforint(r: double): Integer;

function
  imolwfor(lon, lat: double; var x, y: double): Integer;

function
  imolwinvint(r: double): Integer;

function
  imolwinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

function
  imolwforint(r: double): Integer;
{ Initialize the Interrupted Mollweide projection
  --------------------------------------------*/
long imolwforint(double r)
/* (I) Radius of the earth (sphere) */}
begin
//* Place parameters in static storage for common use
  RadiusES := r;
//* Initialize central meridians for each of the 6 regions
  lon_center[0] := 1.0471975512; //*   60.0 degrees */
  lon_center[1] := -2.96705972839; //* -170.0 degrees */
  lon_center[2] := -0.523598776; //*  -30.0 degrees */
  lon_center[3] := 1.57079632679; //*   90.0 degrees */
  lon_center[4] := -2.44346095279; //* -140.0 degrees */
  lon_center[5] := -0.34906585; //*  -20.0 degrees */

//* Initialize false eastings for each of the 6 regions
  feast[0] := RadiusES * -2.19988776387;
  feast[1] := RadiusES * -0.15713484;
  feast[2] := RadiusES * 2.04275292359;
  feast[3] := RadiusES * -1.72848324304;
  feast[4] := RadiusES * 0.31426968;
  feast[5] := RadiusES * 2.19988776387;

//* Report parameters to the user
  ptitle('INTERRUPTED MOLLWEIDE EQUAL-AREA');
  radius(r);
  result := (OK);
end;


function
  imolwfor(lon, lat: double; var x, y: double): Integer;
{Interrupted Mollweide forward equations--mapping lat,long to x,y
long imolwfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude */
              /* (I) Latitude */
             /* (O) X projection coordinate */
             /* (O) Y projection coordinate */}
var
{double adjust_lon(double x);
 /* Function to adjust longitude to -180 - 180 */}
  delta_lon, //* Delta longitude (Given longitude - center */
    theta,
    delta_theta,
    con: Double;
  i, region: Integer;
begin
//* Forward equations  -----------------*/
//* Note:  PI has been adjusted so that the correct region will be assigned
//   when lon = 180 deg.
//   ----------------------------------------------------------------------*/
  if (lat >= 0.0) then
  begin
    if ((lon >= 0.34906585) and (lon < 1.91986217719)) then
      region := 0
    else if
      (((lon >= 1.919862177) and (lon <= (PI + 1.0E-14))) or
      ((lon >= (-PI - 1.0E-14)) and (lon < -1.745329252))) then
      region := 1
    else region := 2;
  end
  else
  begin
    if ((lon >= 0.34906585) and (lon < 2.44346095279)) then
      region := 3
    else if
      (((lon >= 2.44346095279) and (lon <= (PI + 1.0E-14))) or
      ((lon >= (-PI - 1.0E-14)) and (lon < -1.2217304764))) then
      region := 4
    else region := 5;
  end;

  delta_lon := AdjustLon(lon - lon_center[region]);
  theta := lat;
  con := PI * sin(lat);

//* Iterate using the Newton-Raphson method to find theta
 // -----------------------------------------------------*/
  for i := 0 to 50 do
  begin
    delta_theta := -(theta + sin(theta) - con) / (1.0 + cos(theta));
    theta := theta + delta_theta;
    if (abs(delta_theta) < EPSLN) then break;
    //* if (i >= 50) perror("Iteration failed to converge","IntMoll-forward"); */
    if (i >= 50) then
      {PError('Iteration failed to converge', 'IntMoll-forward');}
      DoProjError(241,241,0,0,0);
  end;
  theta := theta / 2.0;

//* If the latitude is 90 deg, force the x coordinate to be "0 + false easting"
//   this is done here because of percision problems with "cos(theta)"
//   --------------------------------------------------------------------------*/
  if (PI / 2 - abs(lat) < EPSLN) then delta_lon := 0;
  x := feast[region] + 0.900316316158 * RadiusES * delta_lon *
    cos(theta);
  y := RadiusES * 1.4142135623731 * sin(theta);
  result := (OK);
end;


function
  imolwinvint(r: double): Integer;
{Initialize the Interrupted Mollweide projection
long imolwinvint(double r)
/* (I) Radius of the earth (sphere) */}
begin
//* Place parameters in static storage for common use
  RadiusES := r;

//* Initialize central meridians for each of the 6 regions
  lon_center[0] := 1.0471975512; //*   60.0 degrees */
  lon_center[1] := -2.96705972839; //* -170.0 degrees */
  lon_center[2] := -0.523598776; //*  -30.0 degrees */
  lon_center[3] := 1.57079632679; //*   90.0 degrees */
  lon_center[4] := -2.44346095279; //* -140.0 degrees */
  lon_center[5] := -0.34906585; //*  -20.0 degrees */

//* Initialize false eastings for each of the 6 regions
  feast[0] := RadiusES * -2.19988776387;
  feast[1] := RadiusES * -0.15713484;
  feast[2] := RadiusES * 2.04275292359;
  feast[3] := RadiusES * -1.72848324304;
  feast[4] := RadiusES * 0.31426968;
  feast[5] := RadiusES * 2.19988776387;

//* Report parameters to the user
  ptitle('INTERRUPTED MOLLWEIDE EQUAL-AREA');
  radius(r);
  result := (OK);
end;


function
  imolwinv(x, y: double; var lon, lat: double): Integer;
{long imolwinv(double x, double y, double *lon, double *lat)
           /* (I) X projection coordinate */
           /* (I) Y projection coordinate */
              /* (O) Longitude */
              /* (O) Latitude */}
var
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180 */}
  theta { temp}: Double;
  region: Integer;
begin
  result := (OK);
//* Inverse equations  -----------------*/
  if (y >= 0.0) then
  begin
    if (x <= RadiusES * -1.41421356248) then region := 0
    else if (x <= RadiusES * 0.942809042) then region := 1
    else region := 2;
  end
  else
  begin
    if (x <= RadiusES * -0.942809042) then region := 3
    else if (x <= RadiusES * 1.41421356248) then region := 4
    else region := 5;
  end;
  x := x - feast[region];

  theta := arcsin(y / (1.4142135623731 * RadiusES));
  lon := AdjustLon(lon_center[region] + (x / (0.900316316158 *
    RadiusES * cos(theta))));
  lat := arcsin((2.0 * theta + sin(2.0 * theta)) / PI);

//* Are we in a interrupted area?  If so,  result:=  status code of IN_BREAK.
//  ---------------------------------------------------------------------*/
  if ((region = 0) and ((lon < 0.34906585) or (lon > 1.91986217719)))
    then result := (IN_BREAK);
  if ((region = 1) and (((lon < 1.91986217719) and (lon > 0.34906585))
    or
    ((lon > -1.74532925199) and (lon < 0.34906585)))) then
      result := (IN_BREAK);
  if ((region = 2) and ((lon < -1.745329252) or (lon > 0.34906585)))
    then result := (IN_BREAK);
  if ((region = 3) and ((lon < 0.34906585) or (lon > 2.44346095279)))
    then result := (IN_BREAK);
  if ((region = 4) and (((lon < 2.44346095279) and (lon > 0.34906585))
    or
    ((lon > -1.2217304764) and (lon < 0.34906585)))) then
      result := (IN_BREAK);
  if ((region = 5) and ((lon < -1.2217304764) or (lon > 0.34906585)))
    then result := (IN_BREAK);

end;


end.
