{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzSter;

{NAME                             STEREOGRAPHIC
PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Stereographic projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.
PROGRAMMER              DATE
----------              ----
T. Mittan		Mar, 1993

ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.

2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface

uses Math;

//* Variables common to all subroutines in this code file
var
  r_major, //* major axis
    lon_center, //* Center longitude (projection center)
    lat_origin, //* center latitude
    false_northing, //* y offset in meters
    false_easting, //* x offset in meters
    sin_p10, //* sin of center latitude
    cos_p10: Double; //* cos of center latitude

function
  sterforint(r_maj, center_lon, center_lat,
  false_east, false_north: double): Integer;

function
  sterfor(lon, lat: double; var x, y: double): Integer;

function
  sterinvint(r_maj, center_lon, center_lat,
  false_east, false_north: double): Integer;

function
  sterinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

function
  sterforint(r_maj, center_lon, center_lat,
  false_east, false_north: double): Integer;
{Initialize the Stereographic projection
long sterforint(double r_maj, double center_lon,
double center_lat, double false_east, double false_north)
                /* major axis
                    /* center longitude
                    /* center latitude
                    /* x offset in meters
                     /* y offset in meters}
  {var temp:Double;}//* temporary variable
begin
//* Place parameters in static storage for common use

  r_major := r_maj;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;
    sin_p10 := Sin(center_lat);
    cos_p10 := Cos(center_lat);
  {sincos(center_lat, sin_p10, cos_p10);}

//* Report parameters to the user

  ptitle('STEREOGRAPHIC');
  radius(r_major);
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  sterfor(lon, lat: double; var x, y: double): Integer;
{Stereographic forward equations--mapping lat,long to x,y
long sterfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude 		*/
              /* (I) Latitude 		*/
             /* (O) X projection coordinate 	*/
             /* (O) Y projection coordinate 	*/}
var
{double AdjustLon(double x);
/* Function to adjust longitude to -180 - 180 	*/}
  sinphi, cosphi, //* sin and cos value				*/
    dlon, //* delta longitude value			*/
    coslon, //* cos of longitude				*/
    ksp, //* scale factor					*/
    g: double;
begin
//* Forward equations

  dlon := AdjustLon(lon - lon_center);
    sinphi := Sin(lat);
    cosphi := Cos(lat);
  {sincos(lat, sinphi, cosphi);}
  coslon := cos(dlon);
  g := sin_p10 * sinphi + cos_p10 * cosphi * coslon;
  if (abs(g + 1.0) <= EPSLN) then
  begin
    {PError('Point projects into infinity', 'ster-for');}
    DoProjError(103,103,0,0,0);
    result := (103);
    exit;
  end
  else
  begin
    ksp := 2.0 / (1.0 + g);
    x := false_easting + r_major * ksp * cosphi * sin(dlon);
    y := false_northing + r_major * ksp * (cos_p10 * sinphi - sin_p10
      *
      cosphi * coslon);
  end;
  result := (OK);
end;


function
  sterinvint(r_maj, center_lon, center_lat,
  false_east, false_north: double): Integer;
{Initialize the Stereographic projection
long sterinvint(double r_maj, double center_lon,
double center_lat, double false_east, double false_north)
                /* major axis			*/
                    /* center longitude		*/
                    /* center latitude		*/
                    /* x offset in meters		*/
                     /* y offset in meters		*/}
{var
 temp:double; }//* temporary variable		*/
{double e0fn(double x),e1fn(double x),e2fn(double x),e3fn(double x);
   /* functions	*/                                 }
begin
//* Place parameters in static storage for common use

  r_major := r_maj;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;
    sin_p10 := Sin(center_lat);
    cos_p10 := Cos(center_lat);
  {sincos(center_lat, sin_p10, cos_p10); }

// Report parameters to the user

  ptitle('STEREOGRAPHIC');
  radius(r_major);
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;

function
  sterinv(x, y: double; var lon, lat: double): Integer;
{ Stereographic inverse equations--mapping x,y to lat/long
long sterinv(double x, double y, double *lon, double *lat)
            /* (O) X projection coordinate 	*/
            /* (O) Y projection coordinate 	*/
               /* (I) Longitude 		*/
               /* (I) Latitude 		*/}
var
{double AdjustLon(double x);
 /* Function to adjust longitude to -180 - 180
double msfnz(double eccent, double sinphi, double cosphi);
  /* function to compute m}
  rh, //* height above ellipsoid
    z, //* angle
    sinz, cosz, //* sin of z and cos of z
    con: double;
begin
//* Inverse equations

  x := x - false_easting;
  y := y - false_northing;
  rh := sqrt(x * x + y * y);
  z := 2.0 * arctan(rh / (2.0 * r_major));
    sinz := Sin(z);
    cosz := Cos(z);
  {sincos(z, sinz, cosz);}
  lon := lon_center;
  if (abs(rh) <= EPSLN) then
  begin
    lat := lat_origin;
    result := (OK);
    exit;
  end
  else
  begin
    lat := arcsin(cosz * sin_p10 + (y * sinz * cos_p10) / rh);
    con := abs(lat_origin) - HALF_PI;
    if (abs(con) <= EPSLN) then
    begin
      if (lat_origin >= 0.0) then
      begin
        lon := AdjustLon(lon_center + arctan(x / -y));
        result := (OK);
        exit;
      end
      else
      begin
        lon := AdjustLon(lon_center - arctan(-x / y));
        result := (OK);
        exit;
      end
    end
    else
    begin
      con := cosz - sin_p10 * sin(lat);
      if ((abs(con) < EPSLN) and (abs(x) < EPSLN)) then
      begin result := (OK); exit; end
      else
        lon := AdjustLon(lon_center + arctan((x * sinz * cos_p10) /
          (con * rh)));
    end;
  end;
  result := (OK);
end;



end.
