{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzTm;

{***********************************************************************
NAME                            TRANSVERSE MERCATOR
PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Transverse Mercator projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.
PROGRAMMER              DATE
----------              ----
D. Steinwand, EROS      Nov, 1991
T. Mittan		Mar, 1993
ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
*******************************************************************************/
#include <stdlib.h>    /* for abs() declaration */
#include "gctpc/cproj.h"
#include "gctpc/report.h"
}

interface

uses Math{,Dialogs,Sysutils};

function
  utmforint(r_maj, r_min, scale_fact: double; zone: Integer):
    Integer;

function
  utmfor(lon, lat: double; var x, y: double): Integer;

function
  utminvint(r_maj, r_min, scale_fact: double; zone: Integer):
    Integer;

function
  utminv(x, y: double; var lon, lat: double): Integer;

function
  tmforint(r_maj, r_min, scale_fact, center_lon,
  center_lat, false_east, false_north: double): Integer;

function
  tmfor(lon, lat: double; var x, y: double): Integer;

function
  tminvint(r_maj, r_min, scale_fact, center_lon,
  center_lat, false_east, false_north: double): Integer;

function
  tminv(x, y: double; var lon, lat: double): Integer;


implementation

uses buzGlobal;

//* Variables common to all subroutines in this code file
var
 {static double} r_major, //* major axis
 {static double} r_minor, //* minor axis
{static double} scale_factor, //* scale factor
{static double} lon_center, //* Center longitude (projection center)
{static double} lat_origin, //* center latitude
{static double} e0, e1, e2, e3, //* eccentricity constants
 {static double} e, es, esp, //* eccentricity constants
 {static double} ml0, //* small value m
{static double} false_northing, //* y offset in meters
{static double} false_easting, //* x offset in meters
 {static double} ind: Double; //* spherical flag


function
  utmforint(r_maj, r_min, scale_fact: double; zone: Integer):
    Integer;
{ Initialize the Universal Transverse Mercator (UTM) projection
long utmforint(
double r_maj,/* major axis
double r_min,/* minor axis
double scale_fact,/* scale factor
long int zone)/* zone number}
var temp: double; //* temporary variable
{double e0fn(double x),e1fn(double x),
e2fn(double x),e3fn(double x),
mlfn(double e0, double e1, double e2, double e3, double phi);
  /* functions		*/}
begin
  if ((abs(zone) < 1) or (abs(zone) > 60)) then
  begin
    {PError('Illegal zone number', 'utm-forint');}
    DoProjError(11,11,0,0,0);
    result := (11);
    Exit;
  end;
{if (zone == 0)   zone = 1;}
  r_major := r_maj;
  r_minor := r_min;
  scale_factor := scale_fact;
  lat_origin := 0.0;
  lon_center := ((6 * abs(zone)) - 183) * D2R;
  false_easting := 500000.0;

  if (zone < 0) then false_northing := 10000000.0
  else false_northing := 0.0;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
  e0 := e0fn(es);
  e1 := e1fn(es);
  e2 := e2fn(es);
  e3 := e3fn(es);
  ml0 := r_major * mlfn(e0, e1, e2, e3, lat_origin);
  esp := es / (1.0 - es);

  if (es < 0.00001) then ind := 1;

//* Report parameters to the user
  ptitle('UNIVERSAL TRANSVERSE MERCATOR (UTM)');
  GenrptLong(zone, 'Zone:     ');
  radius2(r_major, r_minor);
  genrpt(scale_factor, 'Scale Factor at C. Meridian:     ');
  cenlonmer(lon_center);
  result := (OK);
end;


function
  tmforint(r_maj, r_min, scale_fact, center_lon,
  center_lat, false_east, false_north: double): Integer;
{Initialize the Transverse Mercator (TM) projection
long tmforint(
double r_maj,/* major axis
double r_min,/* minor axis
double scale_fact,/* scale factor
double center_lon,/* center longitude
double center_lat,/* center latitude
double false_east,/* x offset in meters
double false_north)/* y offset in meters}
var
  temp: double; //* temporary variable
{double e0fn(double x),e1fn(double x),
e2fn(double x),e3fn(double x),
mlfn(double e0, double e1, double e2, double e3, double phi);
 /* functions	*/}
begin
//* Place parameters in static storage for common use
  r_major := r_maj;
  r_minor := r_min;
  scale_factor := scale_fact;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
  e0 := e0fn(es);
  e1 := e1fn(es);
  e2 := e2fn(es);
  e3 := e3fn(es);
  ml0 := r_major * mlfn(e0, e1, e2, e3, lat_origin);
  esp := es / (1.0 - es);

  if (es < 0.00001) then ind := 1
  else ind := 0;

//* Report parameters to the user
  ptitle('TRANSVERSE MERCATOR (TM)');
  radius2(r_major, r_minor);
  genrpt(scale_factor, 'Scale Factor at C. Meridian:    ');
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  tmfor(lon, lat: double; var x, y: double): Integer;
{ Transverse Mercator forward equations--mapping lat,long to x,y
long tmfor(
double lon,/* (I) Longitude
double lat,/* (I) Latitude
double *x,/* (O) X projection coordinate
double *y)/* (O) Y projection coordinate 	}
var
{double adjust_lon(double x);
 Function to adjust longitude to -180 - 180}
  delta_lon, //* Delta longitude (Given longitude - center
{ theta,	}//* angle
{ delta_theta, }//* adjusted longitude
{double mlfn(double e0, double e1, double e2, double e3, double phi);
   function to compute small m }
  sin_phi, cos_phi, //* sin and cos value
    al, als, //* temporary values
    b, //* temporary values
    c, t, tq, //* temporary values
    con, n, ml: double; //* cone constant, small m
begin
//* Forward equations  -----------------*/
  delta_lon := AdjustLon(lon - lon_center);
  sin_phi := Sin(lat);
  cos_phi := Cos(lat);
    {sincos(lat, sin_phi, cos_phi);}
//* This part was in the fortran code and is for the spherical form
  if (ind <> 0) then
  begin
    b := cos_phi * sin(delta_lon);
    if ((abs(abs(b) - 1.0)) < 0.0000000001) then
    begin
      {PError('Point projects into infinity', 'tm-for');}
      DoProjError(93,93,0,0,0);
      result := (93);
      Exit;
    end
    else
    begin {log((1.0 + b)/(1.0 - b))}
      x := 0.5 * r_major * scale_factor * ln((1.0 + b) / (1.0 - b));
      con := arccos(cos_phi * cos(delta_lon) / sqrt(1.0 - b * b));
      if (lat < 0) then con := -con;
      y := r_major * scale_factor * (con - lat_origin);
      result := (OK);
      Exit;
    end;
  end;

  al := cos_phi * delta_lon;
  als := SQR(al);
  c := esp * SQR(cos_phi);
  tq := tan(lat);
  t := SQR(tq);
  con := 1.0 - es * SQR(sin_phi);
  n := r_major / sqrt(con);
  ml := r_major * mlfn(e0, e1, e2, e3, lat);
  x := scale_factor * n * al * (1.0 + als / 6.0 * (1.0 - t + c + als
    / 20.0 *
    (5.0 - 18.0 * t + SQR(t) + 72.0 * c - 58.0 * esp))) +
      false_easting;

  y := scale_factor * (ml - ml0 + n * tq * (als * (0.5 + als / 24.0 *
    (5.0 - t + 9.0 * c + 4.0 * SQR(c) + als / 30.0 * (61.0 - 58.0 * t
    + SQR(t) + 600.0 * c - 330.0 * esp))))) + false_northing;

  result := (OK);
end;



function
  utmfor(lon, lat: double; var x, y: double): Integer;
{ Universal Transverse Mercator forward equations--mapping lat,long to x,y
long  utmfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude 		*/
              /* (I) Latitude 		*/
             /* (O) X projection coordinate 	*/
             /* (O) Y projection coordinate 	*/}
begin
  Result := (tmfor(lon, lat, x, y));
end;

function
  utminvint(r_maj, r_min, scale_fact: double; zone: Integer):
    Integer;
{ Initialize the Universal Transverse Mercator (UTM) projection
long utminvint(
double r_maj,/* major axis
double r_min,/* minor axis
double scale_fact,/* scale factor
long int zone)/* zone number}
var temp: double; //* temprorary variables
{double e0fn(double x),e1fn(double x),
e2fn(double x),e3fn(double x),
mlfn(double e0, double e1, double e2, double e3, double phi);
 /* functions		*/}
begin
  if ((abs(zone) < 1) or (abs(zone) > 60)) then
  begin
    {PError('Illegal zone number', 'utm-invint');}
    DoProjError(12,11,0,0,0);
    result := (11);
    Exit;
  end;
{if (zone == 0)   zone = 1;}
  r_major := r_maj;
  r_minor := r_min;
  scale_factor := scale_fact;
  lat_origin := 0.0;
  lon_center := ((6 * abs(zone)) - 183) * D2R;
  false_easting := 500000.0;
  if (zone < 0) then
    false_northing := 10000000.0 else
    false_northing := 0.0;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
  e0 := e0fn(es);
  e1 := e1fn(es);
  e2 := e2fn(es);
  e3 := e3fn(es);
  ml0 := r_major * mlfn(e0, e1, e2, e3, lat_origin);
  esp := es / (1.0 - es);

  if (es < 0.00001) then ind := 1
  else ind := 0;

//* Report parameters to the user
  ptitle('UNIVERSAL TRANSVERSE MERCATOR (UTM)');
  GenrptLong(zone, 'Zone:     ');
  radius2(r_major, r_minor);
  genrpt(scale_factor, 'Scale Factor at C. Meridian:     ');
  cenlonmer(lon_center);
  result := (OK);
end;


function
  tminvint(r_maj, r_min, scale_fact, center_lon,
  center_lat, false_east, false_north: double): Integer;
{Initialize the Transverse Mercator (TM) projection
long tminvint(
double r_maj,/* major axis
double r_min,/* minor axis
double scale_fact,/* scale factor
double center_lon,/* center longitude
double center_lat,/* center latitude
double false_east,/* x offset in meters
double false_north)/* y offset in meters		}
var temp: double; //* temporary variable		*/
{double e0fn(double x),e1fn(double x),
e2fn(double x),e3fn(double x),
mlfn(double e0, double e1, double e2, double e3, double phi);
 /* functions		*/}
begin
//* Place parameters in static storage for common use
  r_major := r_maj;
  r_minor := r_min;
  scale_factor := scale_fact;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
  e0 := e0fn(es);
  e1 := e1fn(es);
  e2 := e2fn(es);
  e3 := e3fn(es);
  ml0 := r_major * mlfn(e0, e1, e2, e3, lat_origin);
  esp := es / (1.0 - es);

  if (es < 0.00001) then ind := 1;

//* Report parameters to the user
  ptitle('TRANSVERSE MERCATOR (TM)');
  radius2(r_major, r_minor);
  genrpt(scale_factor, 'Scale Factor at C. Meridian:    ');
  cenlonmer(lon_center);
  origin(lat_origin);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  tminv(x, y: double; var lon, lat: double): Integer;
{ Transverse Mercator inverse equations--mapping x,y to lat,long
long tminv(
double x,/* (I) X projection coordinate
double y,/* (I) Y projection coordinate
double *lon,/* (O) Longitude
double *lat)/* (O) Latitude}
var
//double adjust_lon(double x);
  //* Function to adjust longitude to -180 - 180
  con, phi, //* temporary angles
    delta_phi, //* difference between longitudes
    sin_phi, cos_phi, tan_phi, //* sin cos and tangent values
    c, cs, t, ts, n, r, d, ds, //* temporary variables
    f, h, g, temp: double; //* temporary variables
  i: Integer; //* counter variable
  max_iter: Integer; //* maximun number of iterations
{double asinz(double con);}
begin
  max_iter := 6;
//* fortran code for spherical form
  if (ind <> 0) then
  begin
    f := exp(x / (r_major * scale_factor));
    g := 0.5 * (f - 1 / f);
    temp := lat_origin + y / (r_major * scale_factor);
    h := cos(temp);
    con := sqrt((1.0 - h * h) / (1.0 + g * g));
    lat := asinz(con);
    if (temp < 0) then lat := -lat;
    if ((g = 0) and (h = 0)) then
    begin
      lon := lon_center;
      result := (OK);
      Exit;
    end
    else
    begin {atan2(g,h)}
      lon := AdjustLon(arctan(g / h) + lon_center);
      result := (OK);
      Exit;
    end;
  end;
//* Inverse equations
  x := x - false_easting;
  y := y - false_northing;

  con := (ml0 + y / scale_factor) / r_major;
  phi := con;
  for i := 0 to max_iter + 1 do
  begin
    delta_phi := ((con + e1 * sin(2.0 * phi) - e2 * sin(4.0 * phi) +
      e3 * sin(6.0 * phi))
      / e0) - phi;
{  delta_phi = ((con + e1 * sin(2.0*phi) - e2 * sin(4.0*phi)) / e0) - phi;}

    phi := phi + delta_phi;
    if (abs(delta_phi) <= EPSLN) then break;
    if (i >= max_iter) then
    begin
      {PError('Latitude failed to converge', 'TM-INVERSE');}
      DoProjError(95,95,0,0,0);
      result := (95);
      Exit;
    end;
  end;
  if (abs(phi) < HALF_PI) then
  begin
  sin_phi := Sin(phi);
  cos_phi := Cos(phi);
    {sincos(phi, sin_phi, cos_phi);}
    tan_phi := tan(phi);
    c := esp * Sqr(cos_phi);
    cs := Sqr(c);
    t := Sqr(tan_phi);
    ts := Sqr(t);
    con := 1.0 - es * Sqr(sin_phi);
    n := r_major / sqrt(con);
    r := n * (1.0 - es) / con;
    d := x / (n * scale_factor);
    ds := Sqr(d);
    lat := phi - (n * tan_phi * ds / r) * (0.5 - ds / 24.0 * (5.0 +
      3.0 * t +
      10.0 * c - 4.0 * cs - 9.0 * esp - ds / 30.0 * (61.0 + 90.0 * t
        +
      298.0 * c + 45.0 * ts - 252.0 * esp - 3.0 * cs)));
    lon := AdjustLon(lon_center + (d * (1.0 - ds / 6.0 * (1.0 + 2.0 *
      t +
      c - ds / 20.0 * (5.0 - 2.0 * c + 28.0 * t - 3.0 * cs + 8.0 * esp
        +
      24.0 * ts))) / cos_phi));
  end
  else
  begin
    lat := HALF_PI * sign(y);
    lon := lon_center;
  end;
  result := (OK);
end;


function
  utminv(x, y: double; var lon, lat: double): Integer;
{ Universal Transverse Mercator inverse equations--mapping x,y to lat,long
long utminv(
double x,/* (I) X projection coordinate
double y,/* (I) Y projection coordinate
double *lon,/* (O) Longitude
double *lat)/* (O) Latitude}
begin
  Result := (tminv(x, y, lon, lat));
end;



end.
